      PROGRAM GLMS
C----------------------------------------------------------------------
C Program to calculate a cubic b-spline for normal, Bernoulli or Poisson
c response data (Y), for one X-variable (X) and one optional covariate (XCOV),
c which may be categorical or continuous. If categorical, the categories must
c be entered as numbers not characters ("1" for category "a", etc).
c The program was written by Dolph Schluter, Zoology Department, University of
C British Columbia, Vancouver, B.C. and Doug Nychka (Statistics Department, North
C Carolina State University, Raleigh N.C.). Its basic purpose is to estimate 
C the GLIM spline for a grid of values of the smoothing parameter (lambda), 
C and find the value that minimizes the generalized cross-validation 
C function or score (gcvf). The grid search for lambda is done in a 
C natural log scale. Data are read freeform from a file (specified along with
c other options in the file "glms.in". No missing values are allowed.
C The program can handle up to NMAX cases and MMAX levels of any categorical
c covariate (these can be reset in the first PARAMETER statement).
c This is version 4.0.
C ---------------------------------------------------------------------
      IMPLICIT DOUBLE PRECISION (A-G,O-Z)
      PARAMETER (NMAX=10000,MMAX=20)
      REAL*8 X(NMAX),Y(NMAX),SNU(NMAX),SY(NMAX)
      REAL*8 LAMB,LMIN,YTONU,NUTOY,STOP,GCVF,OCVF,X1,X2,
     +  TR,step,zincrmax,wtmax
      real*8 xcov(nmax,mmax),beta(mmax),xsave(nmax),syc(nmax),
     *  nufit(nmax),unique(mmax),nucorrect(mmax),tmp
      REAL*4 SYOLD(NMAX),MBOOT(NMAX),VBOOT(NMAX),YOLD(nmax),
     +  ERRVAR,par(20),sycold(nmax)
      integer NMAX,MMAX,N,M,i,j,catcov,maxit,numit,IFLAG,nbackfit
      INTEGER MODEL,IREP(NMAX),NPOINT,IBOOT,INFO,NTOT,icol,npar
      INTEGER*4 savseed,iseed
c     real*8 meandip,vardip,XMIN,XMAX
c     real*4 mean,var
c     INTEGER ICRASH,ILOWER,IUPPER,IONCE,IMODE,MODES(5),HERE
      integer icases(nmax),icompress(nmax),ncompress
      real*8 xcompress(nmax),ycompress(nmax)

      IFLAG=0
      info=0
C     IONCE=0

C Prevents overflows in smths:
      wtmax=10.0d0
      zincrmax=5.0d0
c Number of backfits
      nbackfit=5

c ***** read in parameters from file.in

      read(*,*) npar
      do 10 k= 1,npar
        read(*,*) par(k)
10    continue
      read(*,*) SAVSEED
      if(savseed.eq.0) then
        write(*,*) 'Error: random number seed must be > 0'
        pause
        stop
      endif
      iseed=savseed
C data filename is read in DATIN

      M=par(1)
      if(M.gt.1) then
        write(*,*) 'Error: program can handle only 1 covariate'
        pause
        stop
      elseif(M.lt.0) then
        M=0
      endif

      catcov=par(2)
      if(catcov.ne.0) catcov=1

      nstep=nint(par(3))
      X1=par(4)
      IF(NSTEP.LT.1) THEN
        WRITE(*,*) 
     +   'Warning: number of lambdas (nstep) must be at least 1'
        nstep=1
      ELSEIF(nstep.eq.1) then
        step=0.0
      ELSE
        X2=par(5)
        IF(X1.GE.X2) THEN
          WRITE(*,*) 'Warning: lmin must be smaller than lmax'
          X2=par(4)
          X1=par(5)
        ENDIF
        STEP=(X2-X1)/(NSTEP-1)
      ENDIF

      maxit=nint(par(6))
      stop=par(7)
      model=nint(par(8))
      icol=nint(par(9))

      iboot=par(10)
      if (nstep.gt.1.and.iboot.gt.1) then
        write(*,*) 
     +   'Warning: bootstrap available only for fixed lambda'
        iboot=1
      endif

C Read the data (sorted if necessary, and pooled within unique values
C of X. NPOINT will contain the number of unique values of X. 
      CALL DATIN(X,Y,IREP,NMAX,N,NPOINT,icol,M,xsave)
C Count the total number of observations
      NTOT=0
      DO 90 I=1,NPOINT
        NTOT=NTOT+IREP(I)
90    CONTINUE

C     Check that Y has variance, and that it is appropriately bounded for the model
      CALL VARTEST(Y,NPOINT,I)
      IF (I.EQ.1) THEN          
        WRITE(*,*) 'Error: Y has no variance'
        pause
        STOP
      ENDIF
      if(model.ne.1) then
        do 100 i=1,npoint
          if(y(i).lt.0) then
            write(*,*) 
     *     'Error: Y cannot be negative in binomial and Poisson models'
            pause
            stop
          elseif(model.eq.2.AND.y(i).gt.1) then
            write(*,*)
     *     'Error: Y cannot be greater than 1 in binomial model'
            pause
            stop
          endif
100     continue
      endif

c     Set up XCOV if a covariate is present
      if(M.gt.0) then
        do 120 i=1,npoint
          xcov(i,1)=1
          xcov(i,2)=xsave(i)
120     continue
        if(catcov.eq.0) then
          M=M+1
        else
          call DESIGNX(XCOV,nmax,mmax,npoint,m,unique)
        endif
      endif

c     Compress x to its unique values if a covariate is present, keep for backfit
c     XCOMPRESS will be the unique X-values
c     NCOMPRESS is the number of unique X-values.
c     ICOMPRESS will contain cumulative values of IREP corresponding to unique X-values
c     ICASES will contain the number of cases in X corresponding to each unique X
c     YCOMPRESS is the weighted average of Y-values corresponding to each unique X
      if(M.gt.0) then
        call compressx(x,irep,icases,xcompress,icompress,
     *               nmax,npoint,ncompress)
        call compressy(y,irep,icases,icompress,ycompress,
     *               nmax,npoint,ncompress)
      endif

C ---------------------------- main program ------------------------------    

      write(*,*) 'Results will be written to file GLMS.OUT'
      OPEN (8,FILE='glms.out')

      WRITE(8,*) 'No. of observations =',NTOT
      if(m.eq.0) then
        WRITE(8,*) 'No. of unique X-values =',NPOINT
      else
        WRITE(8,*) 'No. of unique X and covariate values =',NPOINT
        if(catcov.eq.1) then
          write(8,*) 'Levels of the covariate =',(unique(i),i=1,m)
        endif
      endif
      IF(NSTEP.NE.1) THEN
        WRITE(8,*) 'No. of grid steps =    ',NSTEP
        WRITE(8,'(A,2G18.6)') ' Range of lambdas =     ',X1,X2
      ELSE
        WRITE(8,'(A,G18.6)') ' Fixed lambda =    ',X1
      ENDIF
      IF(MODEL.NE.1) THEN
        WRITE(8,*) 'Max no. of iterations =',MAXIT
        WRITE(8,'(A,F9.6)') ' Convergence tolerance =',STOP
      ENDIF
c     WRITE(8,'(A)') '  '

C *** first series of spline fits (no bootstrapping yet) *****

      IF(nstep.ne.1) then
        WRITE(8,'(/A)')
     +   ' Search for minimizer of cross validation function'
        if(M.gt.0) write(8,'(a)') ' Covariate fitted simultaneously'
      endIF 

      if(M.gt.0) then
        IF (MODEL.EQ.1) THEN
          WRITE(8,'(/A)') 
     +  ' ln(lambda)   #parameters      GCV score'
        ELSE
          WRITE(8,'(/A,A)') 
     +  ' ln(lambda)   #parameters      GCV score','    #iter'
        ENDIF
      else
        IF (MODEL.EQ.1) THEN
          WRITE(8,'(/A)') 
     +  ' ln(lambda)   #parameters      GCV score       OCV score '
        ELSE
          WRITE(8,'(/A,A)') 
     +  ' ln(lambda)   #parameters      GCV score       OCV score ',
     +  '    #iter'
        ENDIF
      endif

200   do 215 j=1,nstep
        rlam= (j-1)*step + x1
        lamb= dexp( rlam )

        if(M.eq.0) then
          call glisp(lamb,npoint,x,y,sy,snu,penlty,tr,gcvf,OCVF,
     *      maxit,stop,numit,tstop,MODEL,IREP,INFO,nmax,zincrmax,wtmax)
          IF(INFO.GT.0) THEN
            WRITE(*,'(//A)') 
     +      ' Program crashed (GLM weight is zero)'
            pause
            STOP
          ENDIF
          IF(MODEL.EQ.1) THEN
            write(8,'(f10.5,4x,f10.5,4x,2G16.8)') rlam,tr,gcvf,OCVF
          ELSE
            write(8,'(f10.5,4x,f10.5,4x,2G16.8,i5)')
     +        rlam,tr,gcvf,OCVF,numit
          ENDIF
        else
          call backfit(y,x,xcov,npoint,m,irep,model,sy,syc,beta,
     *                 nufit,snu,catcov,stop,maxit,lamb,nmax,mmax,
     *                 tr,numit,gcvf,ocvf,zincrmax,wtmax,nbackfit,
     *                 xcompress,icompress,ycompress,icases,ncompress)
          IF(MODEL.EQ.1) THEN
            write(8,'(f10.5,4x,f10.5,4x,2G16.8)') rlam,tr,gcvf
          ELSE
            write(8,'(f10.5,4x,f10.5,4x,G16.8,i5)')
     +        rlam,tr,gcvf,numit
          ENDIF
        endif
        if( (gcvf.lt.gmin).or.(j.eq.1)) then
          gmin= gcvf
          lmin= rlam
        endif


C       if( numit.eq.maxit.AND.NUMIT.GE.1) then
C         write(8,*) 'NOTE: maximum number of iterations used'
C       endif

215   continue

c write out results only for simplest case: no grid search and no bootstrapping
      if(nstep.eq.1 .and. iboot.lt.2) then
c       WRITE(8,*) ' '
c       WRITE (8,*) '           Mean       Variance'
c       CALL STATS(MEAN,VAR,X,IREP,NPOINT)
c       WRITE (8,'(A,2E13.5)') ' X    ',MEAN,VAR
c       CALL STATS(MEAN,VAR,Y,IREP,NPOINT)
c       WRITE (8,'(A,2E13.5)') ' Y    ',MEAN,VAR
c       CALL STATS(MEAN,VAR,SY,IREP,NPOINT)
c       WRITE (8,'(A,2E13.5)') ' Yhat ',MEAN,VAR

        if(m.eq.0) then
          WRITE (8,*) ' '
          IF(MODEL.EQ.1) THEN
            WRITE(8,'(A)')
     +   '         Y          X        N        Yhat  '
            DO 345 I=1,NPOINT
              WRITE(8,'(2F12.5,I6,1X,F12.5)') Y(I),X(I),IREP(I),SY(I)
345         CONTINUE
          ELSE
            IF(MODEL.EQ.2) THEN
              WRITE(8,'(A)')
     +   '         Y          X        N        Yhat   ln Yhat/(1-Yhat)'
            ELSEIF(MODEL.EQ.3) THEN
              WRITE(8,'(A)')
     +   '         Y          X        N        Yhat      ln Yhat   '
            ENDIF
            DO 350 I=1,NPOINT
              WRITE(8,'(2F12.5,I6,1X,2F12.5)') 
     +             Y(I),X(I),IREP(I),SY(I),SNU(I)
350         CONTINUE
          ENDIF
        else
C         Write out backfit results
          WRITE (8,*) ''
          if(catcov.eq.1) then
            write(8,*) 'Levels of the covariate =',(unique(i),i=1,m)
            write(8,*) 'Covariate coefficients =',(beta(i),i=1,M)
            write(8,*) ''
            write(8,'(a,a)')
     *      ' Last columns are fitted Yhats for each covariate level,',
     *      ' in order:'
c           syc is Yhat after removing differences between levels of the covariate
            write(8,'(/3a12,a6,2a12,a)') 
     *        ' Y  ','X   ','Covariate','N','Yhat ','YhatC',
     *        '     Fitted values ->'
            do 50 i=1,npoint
              do 55 j=1,m
                nucorrect(j)=nufit(i)+beta(j)
                if(model.ne.1) then
                  nucorrect(j)=nutoy(nucorrect(j),model)
                endif
55            continue
              write(8,'(3f12.5,i6,22f12.5)') 
     *          y(i),x(i),xsave(i),irep(i),sy(i),syc(i),
     *          (nucorrect(j),j=1,m)
50          continue
          else
            write(8,*) 'covariate coefficients (intercept,slope):'
            write(8,*) (beta(i),i=1,M)

c           syc is Yhat after removing differences between levels of the covariate
            write(8,'(/3a12,a6,2a12)') 
     *        ' Y  ','X   ','Covariate','N','Yhat ','YhatC'
            do 70 i=1,npoint
            write(8,'(3f12.5,i6,2f12.5)')
     *          y(i),x(i),xsave(i),irep(i),sy(i),syc(i)
70          continue
          endif
        endif

      endif

C ************* End first set of spline fits ***************


C {{{{{{{{{{ Sequence for bootstrap replications }}}}}}}}}}}}
      if(iboot.gt.1) then
        K=1
        IGEN=0
C       MEANDIP=0.0
C       VARDIP=0.0
        DO 400 I=1,NPOINT
          MBOOT(I)=0.0
          VBOOT(I)=0.0
          YOLD(i)=Y(i)
          syold(i)=sy(i)
          sycold(i)=syc(i)
400     CONTINUE

C [[[[[[[[[[[[[[[[[[[ BEGIN BOOT-LOOP ]]]]]]]]]]]]]]]]]]]]
410     CONTINUE
        CALL REBOOT(Y,SY,SYOLD,IREP,NPOINT,MODEL,IFLAG,ERRVAR,
     +     iseed)
C       Check that the random Y's have variance > 0 (var=0 if I=1):
        CALL VARTEST(Y,NPOINT,I)
        IF(I.EQ.1) THEN
          IGEN=IGEN+1
          IF (IGEN.GT.IBOOT) THEN
            WRITE(*,*) 
     +       'Too many degenerate random samples (> #bootstraps)'
            pause
            STOP
          ENDIF
          GOTO 410
        ENDIF

        if(m.eq.0) then
          call glisp(lamb,npoint,x,y,sy,snu,penlty,tr,gcvf,OCVF,
     *      maxit,stop,numit,tstop,MODEL,IREP,INFO,nmax,zincrmax,wtmax)
          IF(INFO.GT.0) THEN
            WRITE(*,'(//A)') 
     +        ' Program crashed (GLM weight is zero)'
            pause
            STOP
          ENDIF
        else
          call compressy(y,irep,icases,icompress,ycompress,
     *               nmax,npoint,ncompress)
          call backfit(y,x,xcov,npoint,m,irep,model,sy,syc,beta,
     *                 nufit,snu,catcov,stop,maxit,lamb,nmax,mmax,
     *                 tr,numit,gcvf,ocvf,zincrmax,wtmax,nbackfit,
     *                 xcompress,icompress,ycompress,icases,ncompress)

        endif

C       Count modes, if desired
c       IF(IMODE.GT.0) THEN
c         CALL COUNTM(IMODE,X,SY,NPOINT,XMIN,XMAX,MODES,5,I,HERE)
c         IF(I.EQ.1) THEN 
c           MEANDIP=MEANDIP+X(HERE)
c           VARDIP=VARDIP+X(HERE)**2
c         ENDIF
c       ENDIF

C       Accumulate sums of yhat and yhat**2 (nu scale)
C       Use "SYC" instead of SY when there is a covariate
        if(m.eq.0) then
          IF(MODEL.EQ.1) THEN
            DO 440 I=1,NPOINT
              MBOOT(I)=MBOOT(I)+SY(I)
              VBOOT(I)=VBOOT(I)+SY(I)**2
440         CONTINUE
          ELSE
            DO 441 I=1,NPOINT
              MBOOT(I)=MBOOT(I)+SNU(I)
              VBOOT(I)=VBOOT(I)+SNU(I)**2
441         CONTINUE
          ENDIF
        else
          IF(MODEL.EQ.1) THEN
            DO 450 I=1,NPOINT
              MBOOT(I)=MBOOT(I)+SYC(I)
              VBOOT(I)=VBOOT(I)+SYC(I)**2
450         CONTINUE
          ELSE
            DO 451 I=1,NPOINT
              tmp=ytonu(syc(i),model)
              MBOOT(I)=MBOOT(I)+tmp
              VBOOT(I)=VBOOT(I)+tmp**2
451         CONTINUE
          ENDIF
        endif

        if(MOD(k,100).eq.0) WRITE (*,*) K
        K=K+1
        IF(K.LE.IBOOT) GOTO 410

C [[[[[[[[[[[[[[[[[[[ END BOOT-LOOP ]]]]]]]]]]]]]]]]]]]]

C       Calculate mean and sd of yhat or yhatc (nu scale)
        DO 460 I=1,NPOINT
          VBOOT(I)=( VBOOT(I)-(MBOOT(I)**2)/IBOOT )/IBOOT
          IF (VBOOT(I).LT.0) VBOOT(I)=0
          VBOOT(I)=SQRT(VBOOT(I))
          MBOOT(I)=MBOOT(I)/IBOOT
460     CONTINUE
        WRITE(8,'(/A)') ' Summary of bootstrap results'
        WRITE(8,'(/A,I5)') 
     +    ' Number of bootstrap replications =',IBOOT
        if (igen.gt.0) WRITE(8,'(A,I5)/') 
     +    ' Number of times bootstrap sample of Y was degenerate = '
     +    ,IGEN

C       output the results 
c        IF (IMODE.NE.0) THEN
c          WRITE(8,'(/A,2G18.6)') ' Counted modes between:',XMIN,XMAX
c          WRITE(8,'(/A)') ' Number of Modes:    0    1    2    3   >3'
c          WRITE(8,'(A11,6X,5I5)') ' Frequency:',(MODES(I),I=1,5)
c          IF(MODES(2).GT.1) THEN
c            VARDIP=( VARDIP-(MEANDIP**2)/MODES(2) )/MODES(2)
c            IF (VARDIP.LT.0) VARDIP=0
c            VARDIP=SQRT(VARDIP)
c            MEANDIP=MEANDIP/MODES(2)
c            WRITE (8,'(A,G18.6,A,G18.6)') 
c     +       ' Location of one mode at X =',MEANDIP,' SE =',VARDIP
c          ENDIF
c        ENDIF

        if(m.eq.0) then
          WRITE(8,'(/9X,A,A)') 
     +      'Y          X          Yhat     Yhat-1SE',
     +      '    Yhat+1SE'
        else
          write(8,'(5a12)') 'Y  ','X  ','YhatC','YhatC-1SE',
     *      'YhatC+1SE'
        endif

        if(m.eq.0) then
          DO 480, I=1,NPOINT
C           Convert mean and Yhat+1se and Yhat-1se back to y scale
            IF(MODEL.EQ.1) THEN
c             tmp=MBOOT(I)
              X1=SYOLD(I)-VBOOT(I)
              X2=SYOLD(I)+VBOOT(I)
            ELSE
c             tmp=NUTOY(DBLE(MBOOT(I)),MODEL)
              X1=NUTOY( YTONU(DBLE(SYOLD(I)),MODEL)-VBOOT(I), MODEL)
              X2=NUTOY( YTONU(DBLE(SYOLD(I)),MODEL)+VBOOT(I), MODEL)
            ENDIF
            WRITE(8,'(6F12.5)') YOLD(I),X(I),SYOLD(I),X1,X2
480       CONTINUE
        else
          DO 490, I=1,NPOINT
C           Convert mean and Yhat+1se and Yhat-1se back to y scale
            IF(MODEL.EQ.1) THEN
c             tmp=MBOOT(I)
              X1=SYCOLD(I)-VBOOT(I)
              X2=SYCOLD(I)+VBOOT(I)
            ELSE
c             tmp=NUTOY(DBLE(MBOOT(I)),MODEL)
              X1=NUTOY( YTONU(DBLE(SYCOLD(I)),MODEL)-VBOOT(I), MODEL)
              X2=NUTOY( YTONU(DBLE(SYCOLD(I)),MODEL)+VBOOT(I), MODEL)
            ENDIF
            WRITE(8,'(6F12.5)') YOLD(I),X(I),SYCOLD(I),X1,X2
490       CONTINUE
        endif
      endif

C {{{{{{{ End sequence for bootstrap replications }}}}}}}}}}

C ------- End main loop -----------

      CLOSE (8)
999   END

      SUBROUTINE DATIN(X,Y,IREP,NMAX,N,NPOINT,iread,M,xsave)
C Subroutine to read data (Y, X or Y, X, N), sort the data by X 
C if necessary, and count replicates in X (stored in IREP). 
C At most NMAX cases can be read. 
C On exit, NPOINT will be the number of unique X-values, and N the 
C total # cases in file.
      INTEGER NMAX,N,IREP(NMAX),IREAD,M,index(nmax)
      REAL*8 X(NMAX),Y(NMAX),xsave(nmax)
      CHARACTER*50 INFILE
C
      READ (*,'(A50)',ERR=10) INFILE
      goto 11
10    WRITE (*,'(/A)') 
     +  ' Error reading data filename from input file' 
      pause
      STOP
11    OPEN (7,FILE=INFILE)

      IF (IREAD.EQ.1) THEN
        if(m.eq.1) then
          DO 20 I=1,NMAX+1
            READ (7,*,END=50,ERR=40) Y(I),X(I),xsave(i),IREP(I)
20        CONTINUE
        else
          DO 21 I=1,NMAX+1
            READ (7,*,END=50,ERR=40) Y(I),X(I),IREP(I)
21        CONTINUE
        endif
      ELSE
        if(m.eq.1) then
          DO 25 I=1,NMAX+1
            READ (7,*,END=50,ERR=40) Y(I),X(I),xsave(i)
            irep(i)=1
25        CONTINUE
        else
          DO 26 I=1,NMAX+1
            READ (7,*,END=50,ERR=40) Y(I),X(I)
            irep(i)=1
26        CONTINUE
        endif
      ENDIF

30    WRITE (*,*) 
     *  'Error: program cannot handle more than ',nmax,' cases'
      pause
      STOP
40    WRITE (*,'(/A,I4)')
     +  ' Error when reading data from input file, line# = ',I
      IF (IREAD.EQ.1) WRITE (*,'(/A/A)') 
     +  ' Possible remedy:',
     *  ' Program expects an integer in last column (no decimel points)' 
      IF (M.gt.0) WRITE (*,'(/A/A)') 
     +  ' Possible remedy:',
     *  ' Covariate must be a number, cannot be a character' 
      pause
	STOP

50    N=I-1
      WRITE (*,'(/1X,I4,A/)')
     +  N,' cases read from file'

      call organizedata(x,y,xsave,irep,index,nmax,n,npoint,m)

      return
99    END

      subroutine organizedata(x,y,z,irep,index,nmax,n,npoint,m)
C Subroutine sorts data by X (carrying Y, IREP and Z along), and then
C sorts by Z within blocks of equal X values if Z is present (Y and IREP are
C carried along.) Then it groups cases having identical values for X (or
C identical values for both X and Z if Z is present), yielding NPOINT cases.
C This subroutine replaces the sort routine and "DREPS" in the original GLMS program.
      integer nmax,index(nmax),n,irep(nmax),npoint
      real*8 x(nmax),y(nmax),z(nmax)
      integer itmp(n),ier,imin,imax
      real*8 xtmp(n),ytmp(n),ztmp(n),ytot

c     Get index
      call dpsort(x,n,index,1,ier)

c     Sort according to the index
      do 20 i=1,n
        k=index(i)
        xtmp(i)=x(k)
        ytmp(i)=y(k)
        ztmp(i)=z(k)
        itmp(i)=irep(k)
20    continue

c     Copy back to original arrays
      do 30 i=1,n
        x(i)=xtmp(i)
        y(i)=ytmp(i)
        z(i)=ztmp(i)
        irep(i)=itmp(i)
30    continue

c      Sort z within x
      if(m.gt.0) then
        imin=1
        imax=1
40      continue
        if(x(imax+1).eq.x(imin)) then
          imax=imax+1
          if(imax.le.n) goto 40
        else
          if(imax.gt.imin) then
c	      write(*,*) imin,imax
            nrange=imax-imin+1
            call dpsort(z(imin),nrange,index,1,ier)
c	      write(*,*) z(imin)
            do 50 i=1,nrange
              k=index(i)+imin-1
              ztmp(i)=z(k)
              ytmp(i)=y(k)
              itmp(i)=irep(k)
50          continue
            do 60 i=1,nrange
              k=imin+i-1
              z(k)=ztmp(i)
              y(k)=ytmp(i)
              irep(k)=itmp(i)
60          continue
          endif
          imin=imax+1
          imax=imin
          if(imax.le.n) goto 40
        endif
      endif

C Combine cases having the same X-values (and same Z-values if Z is present)
C Check that IREP is at least 1 for each X
      DO 70 I=1,N
        IF (IREP(I).LT.1) IREP(I)=1
70    CONTINUE

      NPOINT=1
      IMAX=0
C Find the next IMIN and IMAX
100   IMIN=IMAX+1
      IMAX=IMIN
      if(m.eq.0) then
        IF(IMAX.LT.N.AND.X(IMIN+1).EQ.X(IMIN)) THEN
          DO 110 I=IMIN+1,N
            IF(X(I).NE.X(IMIN)) THEN
              IMAX=I-1
              GOTO 120
            ELSEIF(I.EQ.N) THEN
              IMAX=N
              GOTO 120
            ENDIF
110       CONTINUE
        ENDIF
      else
        IF(IMAX.LT.N.AND.X(IMIN+1).EQ.X(IMIN).AND.
     *      z(IMIN+1).EQ.z(IMIN)) THEN
          DO 111 I=IMIN+1,N
            IF(X(I).NE.X(IMIN).OR.z(I).NE.z(IMIN)) THEN
              IMAX=I-1
              GOTO 120
            ELSEIF(I.EQ.N) THEN
              IMAX=N
              GOTO 120
            ENDIF
111       CONTINUE
        ENDIF
      endif

120   CONTINUE
      IREP(NPOINT)=IREP(IMIN)
      X(NPOINT)=X(IMIN)
      if(m.ne.0) z(npoint)=z(imin)

      IF (IMIN.LT.IMAX) THEN
        YTOT=DBLE(IREP(IMIN))*Y(IMIN)
        DO 130 I=IMIN+1,IMAX
          IREP(NPOINT)=IREP(NPOINT)+IREP(I)
          YTOT=YTOT+DBLE(IREP(I))*Y(I)
130     CONTINUE
        Y(NPOINT)=YTOT/DBLE(IREP(NPOINT))
      ELSE
        Y(NPOINT)=Y(IMIN)
      ENDIF

C Loop back up to start with the next unique X-value
      IF(IMAX.LT.N) THEN
        NPOINT=NPOINT+1
        GOTO 100
      ENDIF

      return
      end

      SUBROUTINE DPSORT (DX, N, IPERM, KFLAG, IER)
C***BEGIN PROLOGUE  DPSORT
C***PURPOSE  Return the permutation vector generated by sorting a given
C            array and, optionally, rearrange the elements of the array.
C            The array may be sorted in increasing or decreasing order.
C            A slightly modified quicksort algorithm is used.
C***LIBRARY   SLATEC
C***CATEGORY  N6A1B, N6A2B
C***TYPE      DOUBLE PRECISION (SPSORT-S, DPSORT-D, IPSORT-I, HPSORT-H)
C***KEYWORDS  NUMBER SORTING, PASSIVE SORTING, SINGLETON QUICKSORT, SORT
C***AUTHOR  Jones, R. E., (SNLA)
C           Rhoads, G. S., (NBS)
C           Wisniewski, J. A., (SNLA)
C***DESCRIPTION
C
C   DPSORT returns the permutation vector IPERM generated by sorting
C   the array DX and, optionally, rearranges the values in DX.  DX may
C   be sorted in increasing or decreasing order.  A slightly modified
C   quicksort algorithm is used.
C
C   IPERM is such that DX(IPERM(I)) is the Ith value in the
C   rearrangement of DX.  IPERM may be applied to another array by
C   calling IPPERM, SPPERM, DPPERM or HPPERM.
C
C   The main difference between DPSORT and its active sorting equivalent
C   DSORT is that the data are referenced indirectly rather than
C   directly.  Therefore, DPSORT should require approximately twice as
C   long to execute as DSORT.  However, DPSORT is more general.
C
C   Description of Parameters
C      DX - input/output -- double precision array of values to be
C           sorted.  If ABS(KFLAG) = 2, then the values in DX will be
C           rearranged on output; otherwise, they are unchanged.
C      N  - input -- number of values in array DX to be sorted.
C      IPERM - output -- permutation array such that IPERM(I) is the
C              index of the value in the original order of the
C              DX array that is in the Ith location in the sorted
C              order.
C      KFLAG - input -- control parameter:
C            =  2  means return the permutation vector resulting from
C                  sorting DX in increasing order and sort DX also.
C            =  1  means return the permutation vector resulting from
C                  sorting DX in increasing order and do not sort DX.
C            = -1  means return the permutation vector resulting from
C                  sorting DX in decreasing order and do not sort DX.
C            = -2  means return the permutation vector resulting from
C                  sorting DX in decreasing order and sort DX also.
C      IER - output -- error indicator:
C          =  0  if no error,
C          =  1  if N is zero or negative,
C          =  2  if KFLAG is not 2, 1, -1, or -2.
C***REFERENCES  R. C. Singleton, Algorithm 347, An efficient algorithm
C                 for sorting with minimal storage, Communications of
C                 the ACM, 12, 3 (1969), pp. 185-187.
C***ROUTINES CALLED  XERMSG
C***REVISION HISTORY  (YYMMDD)
C   761101  DATE WRITTEN
C   761118  Modified by John A. Wisniewski to use the Singleton
C           quicksort algorithm.
C   870423  Modified by Gregory S. Rhoads for passive sorting with the
C           option for the rearrangement of the original data.
C   890619  Double precision version of SPSORT created by D. W. Lozier.
C   890620  Algorithm for rearranging the data vector corrected by R.
C           Boisvert.
C   890622  Prologue upgraded to Version 4.0 style by D. Lozier.
C   891128  Error when KFLAG.LT.0 and N=1 corrected by R. Boisvert.
C   920507  Modified by M. McClain to revise prologue text.
C   920818  Declarations section rebuilt and code restructured to use
C           IF-THEN-ELSE-ENDIF.  (SMR, WRB)
C***END PROLOGUE  DPSORT
C     .. Scalar Arguments ..
      INTEGER IER, KFLAG, N
C     .. Array Arguments ..
      DOUBLE PRECISION DX(*)
      INTEGER IPERM(*)
C     .. Local Scalars ..
      DOUBLE PRECISION R, TEMP
      INTEGER I, IJ, INDX, INDX0, ISTRT, J, K, KK, L, LM, LMT, M, NN
C     .. Local Arrays ..
      INTEGER IL(21), IU(21)
C     .. External Subroutines ..
c      EXTERNAL XERMSG
C     .. Intrinsic Functions ..
      INTRINSIC ABS, INT
C***FIRST EXECUTABLE STATEMENT  DPSORT
      IER = 0
      NN = N
c      IF (NN .LT. 1) THEN
c         IER = 1
c         CALL XERMSG ('SLATEC', 'DPSORT',
c     +    'The number of values to be sorted, N, is not positive.',
c     +    IER, 1)
c         RETURN
c      ENDIF
C
      KK = ABS(KFLAG)
c      IF (KK.NE.1 .AND. KK.NE.2) THEN
c         IER = 2
c         CALL XERMSG ('SLATEC', 'DPSORT',
c     +    'The sort control parameter, KFLAG, is not 2, 1, -1, or -2.',
c     +    IER, 1)
c         RETURN
c      ENDIF
C
C     Initialize permutation vector
C
      DO 10 I=1,NN
         IPERM(I) = I
   10 CONTINUE
C
C     Return if only one value is to be sorted
C
      IF (NN .EQ. 1) RETURN
C
C     Alter array DX to get decreasing order if needed
C
      IF (KFLAG .LE. -1) THEN
         DO 20 I=1,NN
            DX(I) = -DX(I)
   20    CONTINUE
      ENDIF
C
C     Sort DX only
C
      M = 1
      I = 1
      J = NN
      R = .375D0
C
   30 IF (I .EQ. J) GO TO 80
      IF (R .LE. 0.5898437D0) THEN
         R = R+3.90625D-2
      ELSE
         R = R-0.21875D0
      ENDIF
C
   40 K = I
C
C     Select a central element of the array and save it in location L
C
      IJ = I + INT((J-I)*R)
      LM = IPERM(IJ)
C
C     If first element of array is greater than LM, interchange with LM
C
      IF (DX(IPERM(I)) .GT. DX(LM)) THEN
         IPERM(IJ) = IPERM(I)
         IPERM(I) = LM
         LM = IPERM(IJ)
      ENDIF
      L = J
C
C     If last element of array is less than LM, interchange with LM
C
      IF (DX(IPERM(J)) .LT. DX(LM)) THEN
         IPERM(IJ) = IPERM(J)
         IPERM(J) = LM
         LM = IPERM(IJ)
C
C        If first element of array is greater than LM, interchange
C        with LM
C
         IF (DX(IPERM(I)) .GT. DX(LM)) THEN
            IPERM(IJ) = IPERM(I)
            IPERM(I) = LM
            LM = IPERM(IJ)
         ENDIF
      ENDIF
      GO TO 60
   50 LMT = IPERM(L)
      IPERM(L) = IPERM(K)
      IPERM(K) = LMT
C
C     Find an element in the second half of the array which is smaller
C     than LM
C
   60 L = L-1
      IF (DX(IPERM(L)) .GT. DX(LM)) GO TO 60
C
C     Find an element in the first half of the array which is greater
C     than LM
C
   70 K = K+1
      IF (DX(IPERM(K)) .LT. DX(LM)) GO TO 70
C
C     Interchange these elements
C
      IF (K .LE. L) GO TO 50
C
C     Save upper and lower subscripts of the array yet to be sorted
C
      IF (L-I .GT. J-K) THEN
         IL(M) = I
         IU(M) = L
         I = K
         M = M+1
      ELSE
         IL(M) = K
         IU(M) = J
         J = L
         M = M+1
      ENDIF
      GO TO 90
C
C     Begin again on another portion of the unsorted array
C
   80 M = M-1
      IF (M .EQ. 0) GO TO 120
      I = IL(M)
      J = IU(M)
C
   90 IF (J-I .GE. 1) GO TO 40
      IF (I .EQ. 1) GO TO 30
      I = I-1
C
  100 I = I+1
      IF (I .EQ. J) GO TO 80
      LM = IPERM(I+1)
      IF (DX(IPERM(I)) .LE. DX(LM)) GO TO 100
      K = I
C
  110 IPERM(K+1) = IPERM(K)
      K = K-1
      IF (DX(LM) .LT. DX(IPERM(K))) GO TO 110
      IPERM(K+1) = LM
      GO TO 100
C
C     Clean up
C
  120 IF (KFLAG .LE. -1) THEN
         DO 130 I=1,NN
            DX(I) = -DX(I)
  130    CONTINUE
      ENDIF
C
C     Rearrange the values of DX if desired
C
      IF (KK .EQ. 2) THEN
C
C        Use the IPERM vector as a flag.
C        If IPERM(I) < 0, then the I-th value is in correct location
C
         DO 150 ISTRT=1,NN
            IF (IPERM(ISTRT) .GE. 0) THEN
               INDX = ISTRT
               INDX0 = INDX
               TEMP = DX(ISTRT)
  140          IF (IPERM(INDX) .GT. 0) THEN
                  DX(INDX) = DX(IPERM(INDX))
                  INDX0 = INDX
                  IPERM(INDX) = -IPERM(INDX)
                  INDX = ABS(IPERM(INDX))
                  GO TO 140
               ENDIF
               DX(INDX0) = TEMP
            ENDIF
  150    CONTINUE
C
C        Revert the signs of the IPERM values
C
         DO 160 I=1,NN
            IPERM(I) = -IPERM(I)
  160    CONTINUE
C
      ENDIF
      RETURN
      END


      SUBROUTINE STATS(MEAN,VAR,X,IREP,C)
C Subroutine to compute mean and sample variance of data
C in a REAL*8 array X. IREP is integer*2, and gives the
C frequency of X=X(i). C is the size of arrays X, IREP.
C MEAN and VAR are REAL*4.
      INTEGER C,IREP(C)
      REAL*8 X(C)
      REAL*4 MEAN,VAR,N
      MEAN=0.0
      VAR=0.0
      N=0.0
      DO 10 I=1,C
        MEAN=MEAN+DBLE(IREP(I))*X(I)
        VAR=VAR+ DBLE(IREP(I))*( X(I)**2 )
        N=N+IREP(I)
10    CONTINUE
      VAR=( VAR-(MEAN**2)/N )/(N-1.0)
      MEAN=MEAN/N
      END

      SUBROUTINE VARTEST(Y,N,I)
C Subroutine to check that not all the N Y-values are equal
      INTEGER I,N
      REAL*8 Y(N),Y1
      I=0
      Y1=Y(1)
      DO 10 J=2,N
        IF(Y(J).NE.Y1) GOTO 20
10    CONTINUE
      I=1      
20    RETURN
      END

      SUBROUTINE REBOOT(Y,SY,SYOLD,IREP,NPOINT,MODEL,
     +   IFLAG,ERRVAR,iseed)
C Subroutine to collect a bootstrap sample of Y from the Yhat's
C in SYOLD. IFLAG=0 indicates the first call to this subroutine.
C NPOINT is the number of values of Y, SY and SYOLD. MODEL is set on
C input: MODEL=1 indicates Y are normal, MODEL=2 indicates binomial, and
C MODEL=3 indicates Poisson. In the normal case ERRVAR is the variance
C of the deviations around the estimate SYOLD. Otherwise ERRVAR not used. 
      INTEGER MODEL,NPOINT,IFLAG,IBINOM,IPOISS,IREP(NPOINT),n
      REAL*8 Y(NPOINT),SY(NPOINT)
      REAL*4 ERRVAR,GAUSS,SYOLD(NPOINT)
      INTEGER*4 ISEED
C If IFLAG=0 copy spline estimates Yhat into storage (REAL*4 SYOLD(i)),
C get new seed for random number generator, determine variance of errors
C (normal case only), and set IFLAG=1 to indicate this.
      IF(IFLAG.EQ.0) THEN
C       In normal case, determine variance of errors
        IF(MODEL.EQ.1) THEN
          N=0
          ERRVAR=0.0
          DO 10 I=1,NPOINT
            ERRVAR=ERRVAR+IREP(I)*(SY(I)-Y(I))**2.
            N=N+IREP(I)
10        CONTINUE
          ERRVAR=ERRVAR/N
        ENDIF
        IFLAG=1
      ENDIF
C     Create a bootstrap sample
      IF(MODEL.EQ.1) THEN
        DO 20 I=1,NPOINT
          Y(I)=DBLE(SYOLD(I)+GAUSS(0.,SQRT(ERRVAR/IREP(I)),
     +       ISEED))
20      CONTINUE
      ELSEIF (MODEL.EQ.2) THEN
        DO 40 I=1,NPOINT
          Y(I)=DBLE(IBINOM(IREP(I),SYOLD(I),ISEED))/IREP(I)
40      CONTINUE
      ELSE
        DO 60 I=1,NPOINT
          Y(I)=DBLE(IPOISS(IREP(I)*SYOLD(I),ISEED))/IREP(I)
60      CONTINUE
      ENDIF
      END

      SUBROUTINE COUNTM(JOB,X,Y,N,X1,X2,MODES,M,COUNT,HERE)
C Counts the number of modes (JOB=1) or dips (JOB=2) in Y 
C occurring between x=X1 and x=X2. N is the length of X and Y arrays,
C and M is the dimension of MODES. HERE is the location (I) of the
C final mode (or dip)
      INTEGER JOB,N,M,MODES(M),IUP,IDOWN,COUNT,HERE
      REAL*8 X(N),Y(N),X1,X2

      IUP=0
      IDOWN=0
      COUNT=0
      IF(JOB.EQ.1) THEN
        DO 100 I=1,N-1
          IF (X(I).GE.X1.AND.X(I+1).LE.X2) THEN                            
            IF(IUP.EQ.1.AND.Y(I+1).LT.Y(I)) THEN
              COUNT=COUNT+1
              HERE=I
              IUP=0
            ELSEIF( Y(I+1).GT.Y(I) ) THEN
              IUP=1
            ENDIF
          ENDIF
100     CONTINUE

      ELSE 
        DO 200 I=1,N-1
          IF (X(I).GE.X1.AND.X(I+1).LE.X2) THEN                            
            IF(IDOWN.EQ.1 .AND. Y(I+1).GT.Y(I)) THEN
              COUNT=COUNT+1
              HERE=I
              IDOWN=0
            ELSEIF( Y(I+1).LT.Y(I) ) THEN
              IDOWN=1
            ENDIF
          ENDIF
200     CONTINUE
      ENDIF

      IF(COUNT.LT.M) THEN
        MODES(COUNT+1)=MODES(COUNT+1)+1
      ELSE
        MODES(M)=MODES(M)+1
      ENDIF

      RETURN
      END
c Bernoulli and Poisson link functions, called by GLISP
c If JOB=2 model is Bernoulli, if JOB=3 model is Poisson.

      doubleprecision function ytonu(p,link)
      doubleprecision p
      INTEGER link
      if(p.lt.(1e-7)) p=1e-7
      IF(link.EQ.2) THEN
        if((1-p).lt.(1e-7)) p=1-(1e-7)
        ytonu= dlog(p/(1.0-p))
      ELSE
        if (p.gt.1e7) p=1e7
        YTONU= DLOG(P)
      ENDIF
      return
      end

      doubleprecision function nutoy(h,link)
      doubleprecision h
      INTEGER link
      IF(ABS(h).GT.16) THEN
        if (h.gt.16) then 
          h=16
        else
          h=-16
        endif
      ENDIF
      IF(link.EQ.2) THEN
        nutoy= dexp(h)/(1.0+dexp(h))
      ELSE
        NUTOY=DEXP(H)
      ENDIF
      return
      end

      doubleprecision function phi2(h,link)
      doubleprecision h,temp,nutoy
      integer link
      temp= nutoy(h,link)
c     No need to put bounds on phi2 since nutoy just did it
      if(link.eq.2) then
        phi2= temp*(1.0-temp)
      else
        phi2= temp
      endif
      return
      end

      SUBROUTINE PENLIK(Y,IREP,SNU,PENLTY,JOB,LAMB,NPOINT)
C Subroutine to calculate penalized likelihood for binomial and Poisson
C data. That likelihood is likelihood-lamb*penlty.
C Y is the dependent variable (average of all Y's for a given X)
C IREP is the number of Y values for a given X. 
C SNU are the predicted Y values on the transformed scale 
C (to get yhat use SY=nutoy(SNU).
      INTEGER NPOINT,JOB
      INTEGER IREP(NPOINT)
      REAL*8 Y(NPOINT),SNU(NPOINT),PENLTY,LAMB,LOGLIK

      LOGLIK=0.0D0
      IF (JOB.EQ.2) THEN
C binomial
        DO 10 I=1,NPOINT
          LOGLIK=LOGLIK + 
     +     DBLE(IREP(I)) * ( Y(I)*SNU(I) - DLOG(1.0+DEXP(SNU(I))) )
10      CONTINUE
      ELSE
C poisson
        DO 20 I=1,NPOINT
          LOGLIK=LOGLIK +
     +     DBLE(IREP(I)) * (-DEXP(SNU(I)) + Y(I)*SNU(I) )     
20      CONTINUE
      ENDIF

      LOGLIK= LOGLIK - LAMB * PENLTY
      WRITE(8,*) 'Penalized likelihood = ',LOGLIK

      END

C Subroutine GLISP, written by D. Nychka (Statistics, NCSU) with
C modifications by D. Schluter (Zoology, UBC).
c This a program to compute generalized linear model cubic spline
c
c It is assumed that Y is distributed with log density function :
c
c         y*f(t ) - phi(f(t )) + psi(y)
c
c   i.e. Y has a distribution from an exponential family where the 
c natural parameter is f(t ).
c                         k
c Minimization problem :
c
c  min f in H
c
c         sum [(Y * f(t ) - phi(f(t ) ]  + lambda*J (f) 
c        k=1,n   k     k           k               2            
c
c Algorithm uses an iteratively
c reweighted least squares spline approach.
c The weights are updated each time using the estimates
c from the previous fit.
c
c  notation:
c            nu= natural parameter= f(t )
c                                      k
c
c            phi2(nu) = second derivative of phi at nu
c            nutoy(nu)= expected value of y as a function of nu
c            ytonu(Ey)= value of nu as a function of the 
c                    expected value of y
c
c            nutoy and ytonu are related to the "link" functions
c            in the terminology of Glim
C 
C JOB=1 MEANS Y IS NORMAL, EQUAL ERROR VARIANCE
C JOB=2 MEANS Y IS BERNOULLI
C JOB=3 MEANS Y IS POISSON       
C
c Declarations
      subroutine glisp(lamb,npoint,x,y,sy,snu,penlty,tr,gcvf,OCV,
     *   maxit,stop,numit,tstop,JOB,IREP,INFO,nmax,zincrmax,wtmax) 
      implicit double precision (a-g,o-z)
      INTEGER IREP(NMAX),maxit,numit,job,info,it
      real*8 x(nmax),y(nmax),z(NMAX),wt(NMAX),sy(nmax),prvnu(NMAX)
      real*8 snu(nmax),LEV(NMAX),stop,tstop
      real*8 ytonu,nutoy,nuhat,phi2,lamb,maxdf,GCVF,OCV
      real*8 numax, numin,tmp,zincrmax,wtmax

C Start GLIM in binomial and Poisson cases
      IF(JOB.NE.1) THEN
c       Set up initial estimates for first iteration 
        yhat=0.0
        do 11 i=1,npoint
            yhat=yhat+y(i) 
 11     continue
        yhat=yhat/npoint
        nuhat=ytonu(yhat,JOB)
C       Check for very large or very small nuhat - these will yield
C       zero weights in phi2
        IF(DABS(NUHAT).GT.23.0d0) THEN
          INFO=1
          RETURN
        ENDIF
        diag= phi2(nuhat,JOB)
    
        do 15 i=1,npoint
          prvnu(i)=nuhat
          z(i)=nuhat + (y(i)-yhat)/(diag)
          wt(i)=dsqrt( 1 / ( DBLE(IREP(i))*diag ) )
 15     continue

c Start iterating


      maxdf=0.0
      numit= maxit
      do 500 it=1,maxit
c fit a weighted least squares cubic smoothing spline

      call smths(lamb,npoint,x,z,wt,snu,penlty,tr,LEV,nmax)

C Optional display of the penalized liklihoods for this iteration
C      CALL PENLIK(Y,IREP,SNU,PENLTY,JOB,LAMB,NPOINT)

c compute convergence criterion

          numin=snu(1)
          numax=numin
          maxdf=0.0

          do 25 i=1,npoint
               nuhat= snu(i)

C Check for very large nuhat
               IF(DABS(NUHAT).GT.23.0d0) THEN
                 INFO=1
                 RETURN
               ENDIF

               test=dabs(prvnu(i)-nuhat)
               diag= phi2(nuhat,JOB)

c max and min used in convergence criterion
               if(nuhat.lt.numin) numin=nuhat
               if(nuhat.gt.numax) numax=nuhat
               if(test.gt.maxdf) maxdf=test

c compute psuedo-values and weights for next iteration
               yhat= nutoy(nuhat,JOB)
               tmp=(y(i)-yhat)/diag
               if(tmp.gt.zincrmax) tmp=zincrmax
               if(tmp.lt.-zincrmax) tmp=-zincrmax
               z(i)=nuhat + tmp
               wt(i)=dsqrt( 1 / ( DBLE(IREP(i))*diag ) )
               if(wt(i).gt.wtmax) wt(i)=wtmax
               prvnu(i)=nuhat
 25       continue

           tstop=(maxdf)/(numax-numin)
           if( tstop.lt.stop  ) then 
             numit=it
             goto 550 
           endif

500   continue

C  Normal case by-passes the iterative steps. SMTHS call is like the
C  previous, except SY is substituted for SNU

      ELSE
        DO 510 I=1,NPOINT
          WT(I)=1./DSQRT(DBLE(IREP(I)))
510     CONTINUE
        CALL SMTHS(LAMB,NPOINT,X,Y,WT,SY,PENLTY,TR,LEV,nmax)
      ENDIF
c * exit loop
550   continue

C Optional printing of leverages (diagonal of hat matrix)
C      write(*,*) ' '
C      write(*,*) 'LEV'
C      write(*,*) (lev(i),i=1,npoint)
C      write(*,*) ' '

c compute predicted y (AND GCVF AND OCV ON THE BASIS OF SY)
C GCVF IS THE GENERALIZED CROSS-VALIDATION SCORE
C OCV IS THE ORDINARY CROSS-VALIDATION SCORE 
      gcvf=0.0
      OCV=0.0

      if(job.eq.1) then
        do 44 i=1,npoint
         OCV=OCV + DBLE(IREP(I)) * ( (y(i)-sy(i))/(1-LEV(I)) )**2.
         gcvf=gcvf + DBLE(IREP(I)) *  ( y(i)-sy(i) )**2.
44      continue
      else
        do 45 i=1,npoint
         sy(i)=nutoy(snu(i),JOB)
         diag=phi2(snu(i),job)
         OCV=OCV + (DBLE(IREP(I))/diag)*( (y(i)-sy(i))/(1-LEV(I)) )**2.
         gcvf=gcvf + ( DBLE(IREP(I))/diag ) *  ( y(i)-sy(i) )**2.
 45     continue
      endif
      gcvf=npoint*gcvf/(dble(npoint)-tr)**2.
      OCV=OCV/dble(npoint)

      return
      end 


      subroutine smths(lamb,npoint,x,y,wght,sy,penlty,
     +trace,LEV,nmax)

C In this version (modified by D.S.) LEV is passed to the calling
C program (LEV is computed in subroutine LVRAGE, 
C and contains the diagonal of the hat matrix)

C COMPUTES A CUBIC SMOOTHING SPLINE FIT TO A SET OF DATA GIVEN
C NPOINT(=NUMBER OF DATA VALUES) AND LAMBDA(=VALUE OF
C THE SMOOTHING PARAMETER). THE CODE IS ADAPTED FROM A
C PROGRAM IN DEBOOR,C.(1978).A PRACTICAL GUIDE TO SPLINES.
C SPRINGER-VERLAG : NEW YORK. AN O(NPOINT) ALGORITHM
C SUGGESTED BY HUTCHINSON AND DEHOOG IS USED TO COMPUTE
C LVRAGE VALUES AND CONSTRUCT CONFIDENCE INTERVALS.
c Adapted from Randy Eubank Texas A&M
c
      implicit real*8 (a-g,o-z)
      integer npoint,NMAX,iq1,iq3
      REAL*8 WGHT(NMAX),X(NMAX),Y(NMAX),sy(NMAX)
      REAL*8 A(NMAX,4),V(NMAX,7)
      REAL*8 P,SIXP,SIX1MP,LEV(NMAX),trace,lamb,scal
c     real*8 vlam

C **** lambda rescaled below
c scale change to convert lambda from [0,1] scale to range of the data
      iq1 = int(.25*npoint+ 1)
      iq3 = int(.75*npoint+ 1)
c      scal= (( x(iq3)-x(iq1) )*2)**3
      scal= (( x(iq3)-x(iq1) )*2)
      P=1/(lamb*scal+ 1) 
Cold  P=1/(npoint*lamb+ 1)
      CALL SETUPQ(X,WGHT,Y,NPOINT,V,A(1,4),NMAX)
      CALL CHOLD(P,V,A(1,4),NPOINT,A(1,3),A(1,1),NMAX)

C  COMPUTE PIECEWISE POLYNOMIAL COEFFICIENTS AND THE
C     SUM OF SQUARED ERRORS. THE FITTED VALUES ARE 
C     STORED IN A(.,1). LINEAR,QUADRATIC AND CUBIC
C     COEFFICIENTS ARE STORED IN A(.,2-4). 

      SIX1MP=6.*(1.-P)
      DO 61 I=1,NPOINT
         a(i,1)=Y(I) - SIX1MP*(WGHT(I)**2)*A(I,1)
         sy(i)=a(i,1)
   61   continue
      SIXP=6.*P
      DO 62 I=1,NPOINT
   62    A(I,3)=A(I,3)*SIXP
      NPM1=NPOINT - 1
      DO 63 I=1,NPM1
        A(I,4)=(A(I+1,3)-A(I,3))/V(I,4)
   63    A(I,2)=(A(I+1,1)-A(I,1))/V(I,4)
     *        -(A(I,3)+A(I,4)/3.*V(I,4))/2.*V(I,4)
c Compute the value of the penalty term for the smoothing spline
      penlty=0.0
c      npm1=npoint-1
c      do 50 i=1,npm1
c           a1=sy(i)
c           next=sy(i+1)
c           a2=a(i,2)
c           b=a(i,3)/2.
c           c=a(i,4)/6.
c           d=v(i,4)
c           penlty=penlty+4*(b**2)*d
c     *                      + 12*b*c*(d**2)
c     *                      + 12*(c**2)*(d**3)
c           check=a1+a2*d+b*(d**2)+c*(d**3)
c           if (i.le.5) write(*,*) check,next
c 50   continue
C COMPUTE LVRAGE VALUES ,THE VARIANCE ESTIMATE 
C     SGHAT2 AND CONFIDENCE INTERVALS.
      CALL LVRAGE(NPOINT,V,WGHT,SIX1MP,TR,LEV,NMAX)
C The following 10 lines were commented out by D.S., as they appear to 
C duplicate a computation made in glisp.
C       rss=0.0

C      do 100 k=1,npoint
C          rss= rss + ( ( y(k)-sy(k) )/ wght(k) )**2
C  100 continue

C      if( npoint-tr.gt.0.0) then
C         cvf= rss*npoint/( npoint-tr)**2 
C         vlam=cvf
C      else
C         vlam=0
C      endif

      trace=tr
      return

      END

      SUBROUTINE SETUPQ(X,WGHT,Y,NPOINT,V,QTY,NMAX)
C   PUT DELX=X(.+1)-X(.) INTO V(.,4)
C   PUT THE THREE BANDS OF THE MATRIX Q-TRANSP*D INTO
C     V(.,1-3)
C   PUT THE THREE BANDS OF (D*Q)-TRANSP*(D*Q) AT AND 
C     ABOVE THE DIAGONAL INTO V(.,5-7)
C   HERE Q IS THE TRIDIAGONAL MATRIX OF ORDER (NPOINT
C     -2,NPOINT) THAT SATISFIES Q-TRANSP*T=0 AND WGHT
C     IS THE DIAGONAL MATRIX WHOSE DIAGONAL ENTRIES 
C     ARE THE SQUARE ROOTS OF THE WEIGHTS USED IN THE 
C     PENALIZED LEAST-SQUARES CRITERION
       implicit double precision (a-g,o-z)
       REAL*8 WGHT(NMAX),X(NMAX),y(NMAX)
       REAL*8 QTY(NMAX),V(NMAX,7)
       REAL*8 DIFF,PREV
      INTEGER NPOINT,I,NPM1
      NPM1=NPOINT -1
      V(1,4)=X(2)-X(1)
      DO 11 I=2,NPM1
         V(I,4)=X(I+1) - X(I)
         V(I,1)=WGHT(I-1)/V(I-1,4)
         V(I,2)=-WGHT(I)/V(I,4) - WGHT(I)/V(I-1,4)
   11    V(I,3)=WGHT(I+1)/V(I,4)
      V(NPOINT,1)=0.0
      DO 12 I=2,NPM1
   12    V(I,5)=V(I,1)**2 + V(I,2)**2 + V(I,3)**2
      IF(NPM1 .LT. 3)GO TO 14
      DO 13 I=3,NPM1
   13    V(I-1,6)=V(I-1,2)*V(I,1) + V(I-1,3)*V(I,2)
   14 V(NPM1,6)=0.0
      IF(NPM1 .LT. 4)GO TO 16
      DO 15 I=4,NPM1
   15    V(I-2,7)=V(I-2,3)*V(I,1)
   16 V(NPM1-1,7)=0.0
      V(NPM1,7)=0.0
C  CONSTRUCT Q-TRANSP. * Y IN QTY
      PREV=(Y(2) - Y(1))/V(1,4)
      DO 21 I=2,NPM1
         DIFF=(Y(I+1)-Y(I))/V(I,4)
         QTY(I)=DIFF - PREV
   21    PREV=DIFF
      RETURN
      END
      SUBROUTINE CHOLD(P,V,QTY,NPOINT,U,QU,NMAX)
CONSTRUCTS THE UPPER THREE DIAGONALS IN V(I,J),I=2,
C   NPOINT-1,J=1,3 OF THE MATRIX 6*(1-P)*Q-TRANSP*
C   (D**2)*Q + P*R . (HERE R IS THE MATRIX PROPORTIONAL
C   TO Q-TRANSP*KSUBN*Q , WHERE KSUBN IS THE MATRIX 
C   WITH ELEMENTS K(X(I),X(J)) AND K IS THE USUAL
C   KERNEL ASSOCIATED WITH CUBIC SPLINE SMOOTHING.)
C   THE CHOLESKY DECOMPOSITION OF THIS MATRIX IS COMPUTED
C   AND STORED IN V(.,1-3) AND THE EQUATION SYSTEM FOR 
C   THE QUADRATIC COEFFICIENTS OF THE SPLINE IN ITS 
C   PIECEWISE POLYNOMIAL REPRESENTATION IS SOLVED . THE 
C   SOLUTION IS RETURNED IN U.
       REAL*8 P,QTY(NMAX),QU(NMAX),U(NMAX),V(NMAX,7)
       REAL*8 SIX1MP,TWOP,RATIO,PREV
      INTEGER NPOINT,I,NPM1,NPM2
      NPM1=NPOINT - 1
C  CONSTRUCT 6*(1-P)*Q-TRANSP.*(D**2)*Q + P*R
      SIX1MP=6.*(1. - P)
      TWOP=2.*P
      DO 2 I=2,NPM1
         V(I,1)=SIX1MP*V(I,5) + TWOP*(V(I-1,4)+V(I,4))
         V(I,2)=SIX1MP*V(I,6) + P*V(I,4)
    2    V(I,3)=SIX1MP*V(I,7)
      NPM2=NPOINT - 2
      IF(NPM2 .GE. 2)GO TO 10
      U(1)=0.0
      U(2)=QTY(2)/V(2,1)
      U(3)=0.0
      GO TO 41
C  FACTORIZATION : FACTORIZE THE MATRIX AS L*B-INV*
C     L-TRANSP , WHERE B IS A DIAGONAL MATRIX AND L
C     IS UPPER TRIANGULAR.
   10 DO 20 I=2,NPM2
         RATIO=V(I,2)/V(I,1)
         V(I+1,1)=V(I+1,1) - RATIO*V(I,2)
         V(I+1,2)=V(I+1,2) - RATIO*V(I,3)
         V(I,2)=RATIO
         RATIO=V(I,3)/V(I,1)
         V(I+2,1)=V(I+2,1) - RATIO*V(I,3)
   20    V(I,3)=RATIO
C  FORWARD SUBSTITUTION
      U(1)=0.0
      V(1,3)=0.0
      U(2)=QTY(2)
      DO 30 I=2,NPM2
   30    U(I+1)=QTY(I+1) - V(I,2)*U(I) -V(I-1,3)*U(I-1)
C  BACK SUBSTITUTION
      U(NPOINT)=0.0
      U(NPM1)=U(NPM1)/V(NPM1,1)
      DO 40 I=NPM2,2,-1
   40    U(I)=U(I)/V(I,1) - U(I+1)*V(I,2) - U(I+2)*V(I,3)
C  CONSTRUCT Q*U
   41 PREV=0.0
      DO 50 I=2,NPOINT
         QU(I)=(U(I)-U(I-1))/V(I-1,4)
         QU(I-1)=QU(I) - PREV
   50    PREV=QU(I)
      QU(NPOINT)=-QU(NPOINT)
      RETURN
      END
      SUBROUTINE LVRAGE(NPOINT,V,WGHT,SIX1MP,TR,LEV,NMAX)
CONSTRUCTS THE UPPER THREE DIAGONALS OF (6*(1-P)*
C   Q-TRANSP*(D**2)*Q + P*R)-INV USING THE RECURSION
C   FORMULA IN HUTCHINSON,M.F. AND DEHOOG,F.R.(1985).
C   NUMER. MATH. 47,99-106, AND STORES THEM IN V(.,5-7).
C   THESE ARE USED IN POST AND PRE MULTIPLICATION BY
C   Q-TRANSP AND Q TO OBTAIN THE DIAGONAL ELEMENTS OF
C   THE HAT MATRIX WHICH ARE STORED IN THE VECTOR LEV.
C     THE TRACE OF THE HAT MATRIX IS RETURNED IN TR.
       REAL*8 V(NMAX,7),LEV(NMAX),W1,W2,W3,SIX1MP,TR
       REAL*8 WGHT(NMAX)
      INTEGER NPM1,NPM2,NPM3,NPOINT
      NPM1=NPOINT - 1
      NPM2=NPOINT - 2
      NPM3=NPOINT - 3
C   RECURSION FOR DIAGONALS OF INVERSE MATRIX
      V(NPM1,5)=1/V(NPM1,1)
      V(NPM2,6)=-V(NPM2,2)*V(NPM1,5)
      V(NPM2,5)=(1/V(NPM2,1)) - V(NPM2,6)*V(NPM2,2)
      DO 10 I=NPM3,2,-1
         V(I,7)=-V(I,2)*V(I+1,6) - V(I,3)*V(I+2,5)
         V(I,6)=-V(I,2)*V(I+1,5) - V(I,3)*V(I+1,6)
         V(I,5)=(1/V(I,1))- V(I,2)*V(I,6) - V(I,3)*V(I,7)
   10 CONTINUE
C   POSTMULTIPLY BY (D**2)*Q-TRANSP AND PREMULTIPLY BY Q TO
C     OBTAIN DIAGONALS OF MATRIX PROPORTIONAL TO THE 
C     IDENTITY MINUS THE HAT MATRIX.
      W1=1/V(1,4)
      W2= -1/V(2,4) - 1/V(1,4)
      W3=1/V(2,4)
      V(1,1)=V(2,5)*W1
      V(2,1)=W2*V(2,5) + W3*V(2,6)
      V(2,2)=W2*V(2,6) + W3*V(3,5)
      LEV(1)=1. - (WGHT(1)**2)*SIX1MP*W1*V(1,1)
      LEV(2)=1. - (WGHT(2)**2)*SIX1MP*(W2*V(2,1) + W3*V(2,2))
      TR=LEV(1) + LEV(2)
      DO 20 I=4,NPM1
         W1=1/V(I-2,4)
         W2= -1/V(I-1,4) - 1/V(I-2,4)
         W3=1/V(I-1,4)
         V(I-1,1)=V(I-2,5)*W1 + V(I-2,6)*W2 + V(I-2,7)*W3
         V(I-1,2)=V(I-2,6)*W1 + V(I-1,5)*W2 + V(I-1,6)*W3
         V(I-1,3)=V(I-2,7)*W1 + V(I-1,6)*W2 + V(I,5)*W3
         LEV(I-1)=1. - (WGHT(I-1)**2)*SIX1MP*(W1*V(I-1,1) 
     .             + W2*V(I-1,2) + W3*V(I-1,3))
         TR= TR + LEV(I-1)
   20 CONTINUE
      W1=1/V(NPM2,4)
      W2= -1/V(NPM1,4) - 1/V(NPM2,4)
      W3=1/V(NPM1,4)
      V(NPOINT,1)=V(NPM1,5)*W3
      V(NPM1,1)=V(NPM2,5)*W1 + V(NPM2,6)*W2
      V(NPM1,2)=V(NPM2,6)*W1 + V(NPM1,5)*W2
      LEV(NPM1)=1. - (WGHT(NPM1)**2)*SIX1MP*(W1*V(NPM1,1)
     .             + W2*V(NPM1,2))
      LEV(NPOINT)=1. - (WGHT(NPOINT)**2)*SIX1MP*W3*V(NPOINT,1)
      TR= TR + LEV(NPM1) + LEV(NPOINT)
      RETURN 
      END

      REAL FUNCTION UNIRAN(SEED)
C  Standard version of UNIRAN returns a uniformly distributed
C  real value in the range (0,1). The corresponding random
C  integer is returned as the seed. Function taken from a
C  program in K.J. Marse and S. D. Roberts 1983, Simulation 41
C  p 135-139.
C
C  *** Local declarations.
C
      INTEGER*4 B2E15, B2E16, HI15, HI31, LOW15, LOWPRD, MODLUS,
     1        MULT1, MULT2, OVFLOW, SEED
      DATA MULT1/24112/, MULT2/26143/
      DATA B2E15/32768/, B2E16/65536/, MODLUS/2147483647/
C
C  ******************************************************************
C
C  This routine utilizes a multiplicative congruential generator
C  of the form  SEED(I)=SEED(I-1)*A MOD P  where the multiplier
C  A = 630360016 and the modulus  P = 2**31 - 1.
C
C  The implementation uses a modification of Schrage, L. "A More
C  Portable FORTRAN Random Number Generator", ACM Trans. on Math.
C  Software, Vol 5, No 2, June 1979, p. 132-138.
C
C  In order to prevent the intermediate products from exceeding 31
C  bits, the multiplier must be less than 2**15 and so factors
C  24112*26143=630360016 occur in two applications of the procedure.
C
C  ******************************************************************
C
C  Get the 15 high order bits of SEED.
C
      HI15 = SEED/B2E16
C
C  Get the 16 low order bits of SEED and form low order product.
C
      LOWPRD = (SEED - HI15*B2E16) * MULT1
C
C  Get the 15 high order bits from the low order product, and
C  form the 31 high order bits of full product.
C
      LOW15 = LOWPRD/B2E16
      HI31 = HI15*MULT1 + LOW15
C
C  Get overflow past 31st bit of full product.
C
      OVFLOW = HI31/B2E15
C
C  Reconstruct the result after presubtracting modulus(2**31-1)
C
      SEED = (((LOWPRD - LOW15*B2E16) - MODLUS) + (HI31 - OVFLOW*B2E15)
     1          *B2E16) + OVFLOW
C
C  Add the modulus back in if necessary and repeat the procedure for
C  the second multiplier.
C
      IF(SEED .LT. 0) SEED = SEED + MODLUS
      HI15 = SEED/B2E16
      LOWPRD = (SEED - HI15*B2E16)*MULT2
      LOW15 = LOWPRD/B2E16
      HI31 = HI15*MULT2 + LOW15
      OVFLOW = HI31/B2E15
      SEED = (((LOWPRD - LOW15*B2E16) - MODLUS) + (HI31 - OVFLOW*B2E15)
     1          *B2E16) + OVFLOW
      IF(SEED .LT. 0) SEED = SEED + MODLUS
C
C  Because a 24 bit mantissa cannot distinguish single precision
C  values within 1/2**23 at the upper end of the (0,1) interval,
C  this formula places the 2**31 values in cells of 2**8 to create
C  uniform density in the (0,1) interval.
C
      UNIRAN = (2*(SEED/256)+1)/16777216.0
      RETURN
      END

C Function to return a REAL*4 NORMAL(MEAN,SD) random variable.
C Procedure is that of G. Marsaglia and T. A. Bray (1964; SIAM Review 6:
C 260-264). ISEED is INTEGER*4 seed for the random number generator UNIRAN.
C Seed is changed with each call, to yield a stream of random numbers.
C
      FUNCTION GAUSS(MEAN,SD,ISEED)
      INTEGER NRN
      INTEGER*4 ISEED
      REAL GAUSS,MEAN,SD,X,R1,R2,RNN1,RNN2,S,uniran
      SAVE NRN,RNN2
      IF (NRN.EQ.1) THEN
        GAUSS=MEAN+RNN2*SD
        NRN=0
        GOTO 20
      ENDIF
10    X=UNIRAN(ISEED)
      R1=2*X-1
      X=UNIRAN(ISEED)
      R2=2*X-1
      S=R1*R1+R2*R2
      IF (S.GE.1.or.S.LE.0) GOTO 10
      RNN1=R1*SQRT((-2*LOG(S))/S)
      RNN2=R2*SQRT((-2*LOG(S))/S)
      GAUSS=MEAN+RNN1*SD
      NRN=NRN+1
20    CONTINUE
      END

C Function to return an integer*2 Poisson(XM) random variable,
C where XM is the parameter (mean and variance). Routine is
C taken from W.H. Press et al. 1986, Numerical Recipes, p 207.
C ISEED is integer*4 seed for the random number generator UNIRAN.
      INTEGER FUNCTION IPOISS(XM,ISEED)
      PARAMETER (PI=3.141592654)
      INTEGER*4 iseed
      real uniran
      IF (XM.LT.12.) THEN
        G=EXP(-XM)
        EM=-1.
        T=1.
2       EM=EM+1.
        T=T*UNIRAN(ISEED)
        IF (T.GT.G) GOTO 2
      ELSE
        SQ=SQRT(2.*XM)
        ALXM=ALOG(XM)
        G=XM*ALXM-GAMMLN(XM+1.)
1       Y=TAN(PI*UNIRAN(ISEED))
        EM=SQ*Y+XM
        IF (EM.LT.0.) GOTO 1
        EM=INT(EM)
        T=0.9*(1.+Y**2.)*EXP(EM*ALXM-GAMMLN(EM+1.)-G)
        IF (UNIRAN(ISEED).GT.T) GOTO 1
      ENDIF
      IPOISS=EM
      END

      REAL FUNCTION GAMMLN(XX)
C Returns the value ln[gamma(XX)] for XX > 0.
C The following variables could be made real*8 for extra precision:
      REAL COF(6),STP,HALF,ONE,FPF,X,TMP,SER
      DATA COF,STP/76.18009,-86.50532,24.0141,-1.231740,.120858E-2,
     +            -.536382E-5,2.506628/
      DATA HALF,ONE,FPF/0.5,1.0,5.5/
      X=XX-ONE
      TMP=X+FPF
      TMP=(X+HALF)*LOG(TMP)-TMP
      SER=ONE
      DO 11 J=1,6
        X=X+ONE
        SER=SER+COF(J)/X
11    CONTINUE
      GAMMLN=TMP+LOG(STP*SER)
      END

C INTEGER function to return a binomial(n,p) random variable,
C where n is the number of trials, and p is the parameter.
C This routine is inefficient if N is large and P is small.
C ISEED is integer*4 seed for the random number generator UNIRAN.
      INTEGER FUNCTION IBINOM(N,P,ISEED)
      INTEGER*4 ISEED
      INTEGER N
      REAL P,UNIRAN,X,y,gauss
      x=n*p
      y=n*(1-p)
      if (x.lt.20.or.y.lt.20) then
        IBINOM=0
        DO 10 I=1,N
          X=UNIRAN(ISEED)
          IF (X.LE.P) IBINOM=IBINOM+1
10      CONTINUE
      else
        ibinom=nint(gauss(x,sqrt(p*y),iseed))
        if(ibinom.lt.0) ibinom=0
        if(ibinom.gt.n) ibinom=n
      endif
      END

      SUBROUTINE BACKFIT(y,x,xcov,npoint,m,irep,job,sy,syc,beta,
     *                   nufit,snu,catcov,stop,maxit,lamb,nmax,mmax,
     *                   tr,numit,gcvf,ocv,zincrmax,wtmax,nbackfit,
     *                   xcompress,icompress,ycompress,icases,ncompress)
      IMPLICIT DOUBLE PRECISION (A-G,O-Z)
      INTEGER NMAX, MMAX, N, M, irep(nmax),npoint,nbackfit
      integer i, j, k, job,maxit,numit,catcov
      REAL*8 Y(NMAX), X(NMAX), XCOV(NMAX,MMAX), beta(MMAX)
      real*8 sy(nmax),snu(nmax),wt(nmax),z(nmax),prvnu(nmax),tmp
      real*8 snu1(nmax),snu2(nmax),snuc(nmax),syc(nmax)
      real*8 ytonu,nutoy,phi2,yhat,nuhat,diag,tstop,stop
      real*8 maxdf,numax,numin,test,nufit(nmax)
      real*8 lamb,tr,lev(nmax),penlty,mean,gcvf,ocv,zincrmax,wtmax
      integer icases(nmax),icompress(nmax),ncompress
      real*8 xcompress(nmax),ycompress(nmax),snu2compress(nmax),
     *       sycompress(nmax),snucompress(nmax)

      yhat=0.0
      do 60 i=1,npoint
        yhat=yhat+y(i) 
60    continue
      yhat=yhat/npoint
      if(job.eq.1) then
        nuhat=yhat
      else
        nuhat=ytonu(yhat,job)
      endif

      do 65 i=1,npoint
        sy(i)=yhat
        snu(i)=nuhat
        snu1(i)=0.0d0
        snu2(i)=0.0d0
        snu2compress(i)=0.0d0
65    continue

      do 500 k=1,nbackfit

c       GLIM loop #1  - linear regression on XCOV
c       ============
        if(job.eq.1) then
c       Normal case
          do 100 i=1,npoint
            snuc(i)=snu(i)-snu1(i)
            z(i) = y(i) - snuc(i)
            wt(i)=1./dsqrt(dble(irep(i)))
100       continue
          call regress(XCOV,z,nmax,mmax,M,NPOINT,beta,snu1,wt)
          do 110 i=1,npoint
            snu(i)=snu1(i)+snuc(i)
            sy(i)=snu(i)
110       continue

        else
c         setup initial estimates for glim
          do 115 i=1,npoint
            snuc(i)=snu(i)-snu1(i)
            sy(i)=nutoy(snu(i),job)
            diag= phi2(snu(i),JOB)
            tmp= (y(i)-sy(i))/diag
            if(tmp.gt.zincrmax) tmp=zincrmax
            if(tmp.lt.-zincrmax) tmp=-zincrmax
            z(i)=snu(i)+ tmp
            wt(i)=dsqrt( 1 / ( DBLE(IREP(i))*diag ) )
            if(wt(i).gt.wtmax) wt(i)=wtmax
            prvnu(i)=snu1(i)
115       continue

c         Start iterating
          maxdf=0.0
          numit= maxit
          do 180 it=1,maxit
            call regress(XCOV,z,nmax,mmax,M,NPOINT,beta,snu1,wt)

c           compute convergence criterion
            numin=snu1(1)
            numax=numin
            maxdf=0.0
            do 125 i=1,npoint
              nuhat= snu1(i)
C             Check for very large nuhat
              if(nuhat.lt.-12.0d0) then
                nuhat=-12.0d0
                snu1(i)=-12.0d0
              elseif(nuhat.gt.12.0d0) then
                nuhat=12.0d0
                snu1(i)=12.0d0
              endif
              test=dabs(prvnu(i)-nuhat)
              if(nuhat.lt.numin) numin=nuhat
              if(nuhat.gt.numax) numax=nuhat
              if(test.gt.maxdf) maxdf=test

c             compute psuedo-values and weights for next iteration
              snu(i)=snu1(i)+snuc(i)
              sy(i)= nutoy(snu(i),JOB)
              diag=phi2(snu(i),job)
              tmp= (y(i)-sy(i))/diag
              if(tmp.gt.zincrmax) tmp=zincrmax
              if(tmp.lt.-zincrmax) tmp=-zincrmax
              z(i)=snu1(i) + tmp
              wt(i)=dsqrt( 1 / ( DBLE(IREP(i))*diag ) )
              if(wt(i).gt.wtmax) wt(i)=wtmax
              prvnu(i)=snu1(i)
125         continue
            tstop=(maxdf)/(numax-numin)
            if( tstop.lt.stop  ) then 
              numit=it
              goto 199 
            endif
180       continue

199       continue
        endif
c       ends GLIM loop #1
c       =================

c       GLIM loop #2 - spline on X
c       ============

        if(job.eq.1) then
c       Normal case
          do 200 i=1,npoint
            snuc(i)=snu(i)-snu2(i)
            z(i) = y(i) - snuc(i)
200       continue
c         Take the average of the Z's for unique X, put into YCOMPRESS
          call compressy(z,irep,icases,icompress,ycompress,
     *               nmax,npoint,ncompress)
          do 201 i=1,ncompress
            wt(i)=1./dsqrt(dble(icompress(i)))
201       continue
          call smths(lamb,ncompress,xcompress,ycompress,wt,snu2compress,
     *               penlty,tr,LEV,nmax)
c         Take the compressed "snu" values and put back into uncompressed array
          call expandy(snu2compress,snu2,icases,nmax,npoint,ncompress)
          do 205 i=1,npoint
            snu(i)=snu2(i)+snuc(i)
            sy(i)=snu(i)
205       continue

        else
          do 210 i=1,npoint
            snuc(i)=snu(i)-snu2(i)
            sy(i) = nutoy(snu(i),job)
210       continue

c         Take the average of the SY's per unique X
          call compressy(sy,irep,icases,icompress,sycompress,
     *               nmax,npoint,ncompress)
          do 212 i=1,ncompress
            snucompress(i)=ytonu(sycompress(i),job)
            diag = phi2(snucompress(i),job)
            tmp= (ycompress(i)-sycompress(i))/diag
            if(tmp.gt.zincrmax) tmp=zincrmax
            if(tmp.lt.-zincrmax) tmp=-zincrmax
            z(i) = snucompress(i) + tmp
            wt(i)=dsqrt( 1 / ( DBLE(icompress(i))*diag ) )
            if(wt(i).gt.wtmax) wt(i)=wtmax
            prvnu(i)=snu2compress(i)
212       continue

c         Start iterating
          maxdf=0.0
          numit= maxit
          do 280 it=1,maxit
            call smths(lamb,NCOMPRESS,XCOMPRESS,z,wt,
     *                 snu2compress,penlty,tr,LEV,nmax)
c           compute convergence criterion
            numin=snu2compress(1)
            numax=numin
            maxdf=0.0
            do 225 i=1,ncompress
              nuhat= snu2compress(i)
C             Check for very large nuhat
              if(nuhat.lt.-12.0d0) then
                nuhat=-12.0d0
                snu2compress(i)=-12.0d0
              elseif(nuhat.gt.12.0d0) then
                nuhat=12.0d0
                snu2compress(i)=12.0d0
              endif
              test=dabs(prvnu(i)-nuhat)
              if(nuhat.lt.numin) numin=nuhat
              if(nuhat.gt.numax) numax=nuhat
              if(test.gt.maxdf) maxdf=test
225         continue
            call expandy(snu2compress,snu2,icases,nmax,npoint,ncompress)
            do 230 i=1,npoint
              snu(i)=snu2(i)+snuc(i)
              sy(i)=nutoy(snu(i),job)
230         continue
            tstop=(maxdf)/(numax-numin)
            if( tstop.lt.stop  ) then 
              numit=it
              goto 289 
            endif

c           Compute psuedo-values and weights for next iteration
c           First take the average of the SY's per unique X
            call compressy(sy,irep,icases,icompress,sycompress,
     *                 nmax,npoint,ncompress)
            do 240 i=1,ncompress
              snucompress(i)=ytonu(sycompress(i),job)
              diag = phi2(snucompress(i),job)
              tmp= (ycompress(i)-sycompress(i))/diag
              if(tmp.gt.zincrmax) tmp=zincrmax
              if(tmp.lt.-zincrmax) tmp=-zincrmax
              z(i) = snu2compress(i)+ tmp
              wt(i)=dsqrt( 1 / ( DBLE(icompress(i))*diag ) )
              if(wt(i).gt.wtmax) wt(i)=wtmax
              prvnu(i)=snu2compress(i)
240         continue

280       continue

289       continue
        endif
c       ends GLIM loop #2
c       =================
500   continue

C     How well does yhat predict Y?
C     GCVF IS THE GENERALIZED CROSS-VALIDATION SCORE
C     OCV, the ordinary CV score, can't be computed here because we
c     don't have LEV for the linear part. Can get these, but not if X is different
c     length in the linear and spline parts. So just get GCVF, but don't forget to
c     add M to tr to get the total trace (for parameters in the linear part).
c     NOTE: I didn't adjust the value of tr, I just used tr+M in the formula for
c     GCV score. Hence, tr on output remains the effective #parameters of the
c     spline part only.
      gcvf=0.0
      if(job.eq.1) then
        do 520 i=1,npoint
         gcvf=gcvf + DBLE(IREP(I)) *  ( y(i)-sy(i) )**2.
520     continue
      else
        do 540 i=1,npoint
         gcvf=gcvf + ( DBLE(IREP(I))/diag ) *  ( y(i)-sy(i) )**2.
540     continue
      endif
      gcvf=npoint*gcvf/(dble(npoint)-(tr+DBLE(M)))**2.

C     Compute "SYC", which is Yhat after removing differences between levels
c     of the covariate
      if(catcov.eq.1) then
c       Put a zero in the first element of beta, as we are not interested
c       in the intercept, and can use the zero when calculating fitted values.
c       NUFIT will contain the fitted values for the first level of the covariate
        beta(1)=0.0d0
        CALL DGMATV(xcov,beta,nufit,npoint,M,NMAX)
        do 630 i=1,npoint
          nufit(i)=snu(i)-nufit(i)
630     continue

c       Get mean difference between nuhat and nufit
        mean=0.0d0
        N=0
        do 640 i=1,npoint
          mean=mean+(snu(i)-nufit(i))*irep(i)
          N=N+irep(i)
640     continue
        mean=mean/dble(N)
        do 650 i=1,npoint
          syc(i)=nufit(i) + mean
          if(job.ne.1) then
            tmp=syc(i)
            syc(i)=nutoy(tmp,job)
          endif
650     continue
      else
c       NUFIT will contain the fitted values after effect of the covariate
c       is removed
        mean=0.0d0
        N=0
        do 660 i=1,npoint
          nufit(i)=snu(i)-xcov(i,2)*beta(2)
          mean=mean+(snu(i)-nufit(i))*irep(i)
          N=N+irep(i)
660     continue
        mean=mean/dble(N)
        do 665 i=1,npoint
          syc(i)=nufit(i)+mean
          if(job.ne.1) syc(i)=nutoy(syc(i),job)
665     continue
      endif

      return
      end


      subroutine DESIGNX(XCOV,nmax,mmax,N,M,unique)
      INTEGER NMAX,MMAX,N,M,i,j,isunique
      REAL*8 XCOV(NMAX,MMAX),X(NMAX),unique(mmax)
      do 10 i=1,n
        x(i)=xcov(i,2)
10    continue
      unique(1)=x(1)
c     M=1
      do 20 i=2,n
        isunique=1
        do 21 j=1,M
          if(x(i).eq.unique(j)) then
            isunique=0
            goto 20
          endif
21      continue
        M=M+1
        if(M.gt.mmax) then
          write(*,*) 'Program can''t handle more than',mmax,
     *               'levels for covariate'
          write(*,*) 'First',M-1,'levels:'
          write(*,*) (unique(i),i=1,M-1)
          pause
          stop
        endif
        unique(M)=x(i)
20    continue
c     write(*,*) (unique(i),i=1,M)
      do 100 i=1,n
        do 101 j=2,M
          if(x(i).eq.unique(j)) then
            xcov(i,j)=1
          else
            xcov(i,j)=0
          endif
101     continue
100   continue
      return
      end

      subroutine regress(Xin,Yin,nmax,mmax,M,N,beta,Yhat,wt)
c Subroutine to solve (X'X)*beta = (X'Y) for beta
c and to generate the Yhat's from X*beta
c XTX is (X'X)
c The vector XTXlow contains the lower triangular form of XTX (including diagonal)
c XTY is (X'Y)
c beta is the solution vector
c N is the number of cases
c NMAX is the maximum number of cases
c M is the number of columns of the design matrix X (including intercept)
c MMAX is the maximum number of columns of the design matrix X
c wt is the s.d. of Y's (not the var), so is already the square root.
      IMPLICIT DOUBLE PRECISION (A-G,O-Z)
      INTEGER NMAX, MMAX, N, M, L
      integer i, j, k, ie
      REAL*8 Y(NMAX),X(NMAX,MMAX),XT(MMAX,NMAX),Yhat(NMAX)
      REAL*8 XTX(MMAX,MMAX), XTY(MMAX),Yin(NMAX),Xin(NMAX,MMAX)
      real*8 beta(MMAX), XTXlow(MMAX*(MMAX+1)/2), dummy
      real*8 wt(nmax)

c incorporate weights to formulas. This should just involve
c premultiplying both X and Y by diagonal(1/wt) [i.e., by inv(diagonal(wt))] ? 
c try this as follows:
      do 10 i=1,n
        do 11 j=1,m
          X(i,j)=Xin(i,j)/wt(i)
11      continue
        Y(i)=Yin(i)/wt(i)
10      continue

c Transpose X to yield XT
      call DGTRAN(X,XT,N,M,NMAX,MMAX)
c     do 20 i=1,m
c       write(*,*) (xt(i,j),j=1,n)
c20   continue

C Multiply XT*X to yield XTX
      CALL DGMULT(XT,X,XTX,M,N,M,MMAX,NMAX,MMAX)
c     do 30 i=1,m
c      write(*,*) (xtx(i,j),j=1,m)
c30   continue

c Obtain XTXlow, the lower triangular of XTX
      L=M*(M+1)/2
      k=1
      do 36 j=1,m
        do 35 i=1,j
          XTXlow(k)=XTX(i,j)
          k=k+1
35      continue
36    continue
c     do 38 k=1,L
c       write(*,*) XTXlow(k)
c38   continue

c Multiply XT*Y to yield XTY
      CALL DGMATV(XT,Y,XTY,M,N,MMAX)
c     do 40 i=1,m
c     write(*,*) xty(i)
c40   continue

c Solve for beta
      do 100, i=1,m
        beta(i)=XTY(i)
100   continue
      call dsym(XTXlow,M,beta,ie,0,dummy,*151,*152)
c When doing backfitting, use -M on all but the first backfit to indicate
c that the XTXlow is now the decomposed matrix, no need to decompose each iter
c     call dsym(XTXlow,-M,beta,ie,0,dummy,*151,*152)
      goto 180

151   if (ie.gt.0) then
        write(*,*) 'X''X is not positive-definite'
        write(*,*) 'Row number',ie
        pause
        stop
      elseif(ie.lt.0) then
        write(*,*) 'X''X is too ill-conditioned'
        ie=abs(ie)
        write(*,*) 'Row number',ie
        pause
        stop
      endif

152   write (*,*) 'Error in call to dsym'
      pause
      stop

180   continue
c Multiply Xin*beta to yield Yhat
      CALL DGMATV(Xin,beta,Yhat,N,M,NMAX)

      return
      end

c ------------------------------------------------

      SUBROUTINE DGTRAN(A,B,M,N,NA,NB)
      REAL*8 A(NA,1),B(NB,1)
      DO 1 J=1,M
      DO 1 I=1,N
1     B(I,J)=A(J,I)
      RETURN
      END

      SUBROUTINE DGMULT(A,B,C,M,N,L,NA,NB,NC)
      REAL*8 A(NA,1),B(NB,1),C(NC,1)
      DO 1 J=1,L
      DO 1 I=1,M
      C(I,J)=0.D0
      DO 2 K=1,N
2     C(I,J)=C(I,J)+A(I,K)*B(K,J)
1     CONTINUE
      RETURN
      END

      SUBROUTINE DGMATV(A,V,W,M,N,NDIMA)
      REAL*8 A(NDIMA,1),W(1),V(1)
      DO 1 I=1,M
      W(I)=0.D0
      DO 1 J=1,N
1     W(I)=W(I)+A(I,J)*V(J)
      RETURN
      END

      SUBROUTINE DSYMD(A,N,IE,*)
      IMPLICIT REAL*8 (A-H,O-Z)
C WHERE:
C     A IS A REAL*8 ARRAY DIMENSIONED N*(N+1)/2. ON ENTRY,
C       THE LOWER HALF OF THE A MATRIX BY ROW; ON EXIT,
C       THE LOWER HALF OF THE L MATRIX BY ROW.
C     N IS THE NUMBER OF ROWS
C     IE IS SET ON ERROR RETURN. |IE| IS THE ROW NUMBER IN ERROR.
C       IF IE > 0, THE MATRIX IS NOT POSITIVE DEFINITE; IF IE < 0,
C       THE RATIO OF THE LARGEST TO THE SMALLEST ELEMENT ON THE
C       DIAGONAL OF THE L MATRIX IS LESS THAN OR EQUAL TO 16**(-13).
C       ACTUALLY USES THE SQUARE OF THIS RATIO.
C
      DIMENSION A(1)
      IE=0
      I=1
      IF (N.LE.0) GO TO 80
      IF (A(1).LE.0.0D0) GO TO 90
      A(1)=1.0D0/DSQRT(A(1))
      IF (N.LE.1) GO TO 80
      K1=1
        DO 70 I=2,N
        K1=K1+I-1
        K2=K1+I-1
        LL=1
          DO 60 K=K1,K2
          SUM=0.0D0
          L1=K1
          L2=K-1
          IF (K.EQ.K1) GO TO 20
            DO 10 L=L1,L2
            SUM=SUM+A(L)*A(LL)
            LL=LL+1
10          CONTINUE
          A(K)=A(K)-SUM
          IF (K.EQ.K2) GO TO 30
20        A(K)=A(K)*A(LL)
          LL=LL+1
          GO TO 60
30        TEMP=A(K)
          IF (TEMP.LE.0.0D0) GO TO 90
          IF (TEMP*0.03125D0+SUM.EQ.SUM) GO TO 100
          A(K)=1.0D0/DSQRT(TEMP)
60        CONTINUE
70      CONTINUE
80    RETURN
90    IE=I
      RETURN 1
100   IE=-I
      RETURN 1
      END
      SUBROUTINE DSYMI(A,NN,IE,*)
      IMPLICIT REAL*8 (A-H,O-Z)
C WHERE:
C     A IS A REAL*8 ARRAY DIMENSIONED N*(N+1)/2. ON ENTRY,
C       IT CONTAINS THE LOWER HALF OF A BY ROW; ON EXIT,
C       THE LOWER HALF OF A-INVERSE BY ROW.
C     NN IS THE NUMBER OF ROWS
C       IF NN LESS THAN ZERO DSYMD DOES NOTHING
C     IE IS AS DEFINED BY DSYMD
C
      DIMENSION A(1)
      N=IABS(NN)
C
C     GET CHOLESKY DECOMPOSITION
C
      CALL DSYMD(A,NN,IE,*70)
C
C     CALCULATE L-INVERSE (DIAGONAL ELEMENTS ARE CORRECT)
C
      IF (N.EQ.1) GO TO 35
      K1=1
        DO 30 I=2,N
        K1=K1+I-1
        K2=K1+I-2
        LL=1
        J=1
          DO 20 K=K1,K2
          SUM=0.0D0
          L1=K
          JJ=J
          L2=LL
            DO 10 L=L1,K2
              SUM=SUM+A(L)*A(L2)
            L2=L2+JJ
            JJ=JJ+1
10          CONTINUE
          A(K)=-A(K2+1)*SUM
          J=J+1
          LL=LL+J
20        CONTINUE
30      CONTINUE
C
C     MULTIPLY TRANSPOSE-L-INVERSE AND L-INVERSE FOR A-INVERSE
C
35    NP=N+1
      L1=1
        DO 60 J=1,N
        L=NP-J
        L2=L1
        I2=J
          DO 50 K=1,L
          SUM=0.0D0
          L3=L2
          I3=I2
            DO 40 KK=K,L
            SUM=SUM+A(L3)*A(L3+K-1)
            L3=L3+I3
            I3=I3+1
40          CONTINUE
          A(L2)=SUM
          L2=L2+I2
          I2=I2+1
50        CONTINUE
        L1=L1+J+1
60      CONTINUE
      RETURN
70    RETURN 1
      END
      FUNCTION DSYMS(A,N,B)
      IMPLICIT REAL*8(A-H,O-Z)
C WHERE:
C     DSYMS IS THE LARGEST SOLUTION ELEMENT (ABSOLUTE VALUE)
C     A IS A REAL*8 ARRAY CONTAINING THE L-MATRIX BY ROW.
C     |N| IS THE NUMBER OF ROWS.
C     B IS A REAL*8 ARRAY DIMENSIONED |N|. ON ENTRY, IT
C       CONTAINS A RHS; ON EXIT, A SOLUTION VECTOR.
C
      DIMENSION A(1),B(1)
      NP=IABS(N)
C
C     SOLVE L*Y = B BY FORWARD SUBSTITUTION
C
      B(1)=A(1)*B(1)
      IF (NP.LE.1) GO TO 50
      K1=1
        DO 20 I=2,NP
        K1=K1+I-1
        K2=K1+I-2
        J=1
        SUM=0.0D0
          DO 10 K=K1,K2
          SUM=SUM+A(K)*B(J)
          J=J+1
10        CONTINUE
        B(J)=(B(J)-SUM)*A(K2+1)
20    CONTINUE
C
C     SOLVE  LT*X = Y  BY BACK SUBSTITUTION
C
      NM=NP-1
        DO 40 I=1,NM
        B(J)=B(J)*A(K2+1)
        X=B(J)
        JJ=1
          DO 30 K=K1,K2
          B(JJ)=B(JJ)-A(K)*X
          JJ=JJ+1
30        CONTINUE
        J=J-1
        K2=K1-2
        K1=K1-J
40      CONTINUE
50    B(1)=A(1)*B(1)
C
C     FIND LARGEST ABSOLUTE ELEMENT
C
      DSYMS=0.0D0
        DO 60 I=1,NP
        X=DABS(B(I))
        IF (X.GT.DSYMS) DSYMS=X
60      CONTINUE
      RETURN
      END
      SUBROUTINE DSYM(A,N,B,IE,ITER,CON,*,*)
      IMPLICIT REAL*8 (A-H,O-Z)
C WHERE:
C     A IS A REAL*8 ARRAY DIMENSIONED |N|*(|N|+1)/2.
C       IT CONTAINS EITHER THE A- OR THE L-MATRIX BY ROW.
C     |N| IS THE NUMBER OF ROWS. IF N > 0, A CONTAINS THE
C       ORIGINAL MATRIX; IF N < 0, THE DECOMPOSITION HAS
C       BEEN DONE; IF N = 0, RESET THINGS.
C     B IS A REAL*8 ARRAY. ON ENTRY IT CONTAINS A RHS; ON EXIT,
C       A SOLUTION VECTOR.
C     IE IS A FLAG RETURNED TO THE USER. ON A NORMAL RETURN,
C       IT GIVES THE NUMBER OF ITERATIONS DONE (IF NEGATIVE,
C       AN ESTIMATE OF THE NUMBER OF RQUIRED ITERATIONS).
C       ON RETURN 1, |IE| GIVES ROW NUMBER IN ERROR. IF
C       IE > 0, THE MATRIX IS NOT POSITIVE DEFINITE; IE  < 0,
C       THE MATRIX IS BADLY CONDITIONED.
C     ITER AND CON ARE INCLUDED FOR COMPATIBILITY
C
      DIMENSION A(1),B(1)
      CON=0.0D0
      IF (N.GT.0) CALL DSYMD(A,N,IE,*10)
      IF (N.EQ.0) RETURN
      XBIG=DSYMS(A,N,B)
      IF (ITER.GT.0) RETURN 2
      RETURN
10    RETURN 1
      END

      subroutine compressx(x,irep,icases,
     *                     xcompress,icompress,
     *                     nmax,npoint,ncompress)
c Reduces X (already sorted) to its unique values (placed in Xcompress) by combining cases.
c Icompress will contain the cumulated values of IREP corresponding to unique X-values.
c ICASES records the number of cases in X containing the corresponding unique X-values.
c ncompress is the number of unique X-values.
      integer nmax,irep(nmax),npoint,i,j,k
      real*8 x(nmax),xcompress(nmax)
      integer icases(nmax),icompress(nmax),ncompress
      k=1
      j=1
      i=2
      xcompress(1)=x(1)
      icompress(1)=irep(1)
      icases(1)=1
10    continue
      if(x(i).eq.x(j)) then
        icompress(k)=icompress(k)+irep(i)
        icases(k)=icases(k)+1
        if(i.lt.npoint) then
          i=i+1
          goto 10
        endif
      else
        k=k+1
        xcompress(k)=x(i)
        icompress(k)=irep(i)
        icases(k)=1
        j=i
        if(i.lt.npoint) then
          i=i+1
          goto 10
        endif
      endif
      ncompress=k
      return
      end

      subroutine compressy(y,irep,icases,icompress,ycompress,
     *                     nmax,npoint,ncompress)
c Computes a weighted average of Y-values corresponding to cases having the same X-value.
c Icompress contains the cumulated values of IREP corresponding to unique X-values.
c ICASES contains the number of cases in X containing the corresponding unique X-values.
c ncompress is the number of unique X-values.
      integer nmax,irep(nmax),npoint,i,j,k
      integer icases(nmax),icompress(nmax),ncompress
      real*8 y(nmax),ycompress(nmax),ysum

      k=1
      do 10 i=1,ncompress
        ysum=0.0
        do 11 j=1,icases(i)
          ysum=ysum+dble(irep(k))*y(k)
          k=k+1
11      continue
        ycompress(i)=ysum/icompress(i)
10    continue
      return
      end

      subroutine expandy(snucompress,snu,icases,nmax,npoint,ncompress)
      integer nmax,icases(nmax),ncompress,i,j,k
      real*8 snucompress(nmax),snu(nmax)
      k=1
      do 10 i=1,ncompress
        do 11 j=1,icases(i)
          snu(k)=snucompress(i)
          k=k+1
11      continue
10    continue
      return
      end
